

/**
 Single Author Info:
 username	FirstName 	LastName
 fliu3		Fang		Liu
**/

#ifndef _PI_KERNEL_H_
#define _PI_KERNEL_H_

#include <stdio.h>
#include <stdlib.h>


/* V1: assign four streams to four threads in one block */
__global__ void RF_kernel_v1(double *g_arr_val, int *g_ng, int *g_n, int *g_seed, int len)
{
	__shared__ int ng[4];
	__shared__ int n[4];
	__shared__ int seed[4];

	int tid = threadIdx.x;

	/* each thread stores seed in shared memory */
	ng[tid] = g_ng[tid];
	__syncthreads();
	n[tid] = g_n[tid];
	__syncthreads();
	seed[tid] = g_seed[tid];
	__syncthreads();

	/* do calculation in shared memory and store final values into global memory */	
	for (int i = tid; i < len; i += 4) {
		seed[tid] = (seed[tid] * ng[tid]) % n[tid];	
		__syncthreads();
		g_arr_val[i] = seed[tid];
		__syncthreads();
	}		
}

/* V2: each block has 4 threads while multiple blocks are executing simultaneously */
__global__ void RF_kernel_v2(double *g_arr_val, int *g_ng, int *g_n, int *g_seed, int len, int num_blks)
{
        __shared__ int ng[4];
	__shared__ int n[4];
	__shared__ int seed[4];

	int tid = threadIdx.x;
	int idx = blockIdx.x * blockDim.x + threadIdx.x;

	/* each thread stores seed in shared memory */
	ng[tid] = g_ng[tid];
	__syncthreads();
	n[tid] = g_n[tid];
	__syncthreads();
	seed[tid] = g_seed[tid];
	__syncthreads();

	int stride = num_blks * 4;
	__syncthreads();

	/* do calculation in shared memory and store final values into global memory */	
	for (int i = 0; i < blockIdx.x + 1; i++) {
		seed[tid] = (seed[tid] * ng[tid]) % n[tid];	
		__syncthreads();
	}

	for (int j = idx; j < len; j += stride) {
		g_arr_val[j] = seed[tid];
		__syncthreads();

		for (int k = 0; k < num_blks; k++) {
			seed[tid] = (seed[tid] * ng[tid]) % n[tid];	
			__syncthreads();
		}
	}		
}

__global__ void RF_kernel_v3(double *g_arr_val, int *g_ng, int *g_n, int *g_seed, int len)
{
	__shared__ int ng[16];
	__shared__ int n[16];
	__shared__ int seed[16];

	int tid = threadIdx.x;
	
	/* each thread stores seed in shared memory */
	ng[tid] = g_ng[tid % 4];
	__syncthreads();
	n[tid] = g_n[tid % 4];
	__syncthreads();
	seed[tid] = g_seed[tid % 4];
	__syncthreads();

	if (tid < 4) {
		seed[tid] = (seed[tid] * ng[tid]) % n[tid];	
		__syncthreads();
	} else if (tid < 8){
		seed[tid] = (seed[tid] * ng[tid] * ng[tid]) % n[tid];	
		__syncthreads();
	} else if (tid < 12) {
		seed[tid] = (seed[tid] * ng[tid] * ng[tid] * ng[tid]) % n[tid];	
		__syncthreads();
	} else {
		seed[tid] = (seed[tid] * ng[tid] * ng[tid] * ng[tid] * ng[tid]) % n[tid];	
		__syncthreads();
	}

	/* do calculation in shared memory and store final values into global memory */	
	for (int i = tid; i < len; i += 16) {
		g_arr_val[i] = seed[tid];
		__syncthreads();
		seed[tid] = (seed[tid] * ng[tid] * ng[tid] * ng[tid] * ng[tid]) % n[tid];	
		__syncthreads();
	}		
	
}

__global__ void RF_kernel_v4(double *g_arr_val, int *g_ng, int *g_n, int *g_seed, int len, int num_blks)
{
	__shared__ int ng[8];
	__shared__ int n[8];
	__shared__ int seed[8];

	int tid = threadIdx.x;
	int idx = blockIdx.x * blockDim.x + threadIdx.x;

	/* each thread stores seed in shared memory */
	ng[tid] = g_ng[tid % 4];
	__syncthreads();
	n[tid] = g_n[tid % 4];
	__syncthreads();
	seed[tid] = g_seed[tid % 4];
	__syncthreads();

	if (tid < 4) {
		seed[tid] = (seed[tid] * ng[tid]) % n[tid];	
		__syncthreads();
	} else {
		seed[tid] = (seed[tid] * ng[tid] * ng[tid]) % n[tid];	
		__syncthreads();
	}

	for (int i = 0; i < blockIdx.x; i++) {
		seed[tid] = (seed[tid] * ng[tid] * ng[tid]) % n[tid];	
		__syncthreads();
	}
	

	int stride = num_blks * 8;
	/* do calculation in shared memory and store final values into global memory */	
	for (int j = idx; j < len; j += stride) {
		g_arr_val[j] = seed[tid];
		__syncthreads();

		for (int k = 0; k < num_blks; k++) {
			seed[tid] = (seed[tid] * ng[tid] * ng[tid]) % n[tid];	
			__syncthreads();
		}
	}		
}

#endif
